-- Lua optimizations
local date = date
local format = format
local gsub = gsub
local strfind = strfind
local strlen = strlen
local strsplit = strsplit
local strsub = strsub
local strupper = strupper
local time = time
local tonumber = tonumber
local GetRealmName = GetRealmName
local GetNumSavedInstances = GetNumSavedInstances
local GetSavedInstanceInfo = GetSavedInstanceInfo
local SecondsToTime = SecondsToTime
local RequestRaidInfo = RequestRaidInfo
local UnitClass = UnitClass
local UnitName = UnitName

-- Use Blizzard's strings to save localization efforts
local NO_RAID_INSTANCES_SAVED = NO_RAID_INSTANCES_SAVED -- "You are not saved to any instances"
local NAME = NAME -- "Name"
local EMBLEM_SYMBOL = EMBLEM_SYMBOL -- "Icon"
local TIMEMANAGER_24HOURMODE = TIMEMANAGER_24HOURMODE -- "24 Hour Mode"
local IGNORE_LIST = IGNORE_LIST -- "Ignore List"

local SavedInstances = LibStub("AceAddon-3.0"):NewAddon("SavedInstances", "AceEvent-3.0")
_G.SavedInstances = SavedInstances
local Config = LibStub("AceConfig-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("SavedInstances")

local LDB = LibStub("LibDataBroker-1.1", true)
local LDBIcon = LDB and LibStub("LibDBIcon-1.0", true)
local launcher

local QTip = LibStub("LibQTip-1.0")
local tooltip

local db
local defaultDB = {
	Toons = { }, 	-- table key: "Toon - Realm"; value:
				-- Class: string
				-- Hide: boolean
				-- NumSaved: integer
	Tooltip = {
		FreshToons = false,
		AllRaids = true,
		Details = false,
	},
	Instances = { }, 	-- table key: "Instance name"; value:
					-- Show: boolean
					-- Order: integer
					-- table key: "Toon - Realm"; value:
						-- table key: "Difficulty"; value:
							-- ID: integer
							-- Expires: integer
	FuBar = {
		AttachMinimap = false,
		HideMinimapButton = true,
	},
	MinimapIcon = { },
	profile = { },	-- LibFuBarPlugin-3.0 expects this /sigh
}

function SavedInstances:OnInitialize()
	local toon = UnitName("player").." - "..GetRealmName()
	SavedInstancesDB = SavedInstancesDB or defaultDB
	self.db = db or SavedInstancesDB
	self.db.Toons[toon] = self.db.Toons[toon] or { }
	self.db.Toons[toon].Class = self.db.Toons[toon].Class or select(2, UnitClass("toon"))
	self.db.Toons[toon].Hide = self.db.Toons[toon] or false
	self.db.Toons[toon].NumSaved = self.db.Toons[toon].NumSaved or 0
	SavedInstances:SetupOptions()
	if LDB then
		self.launcher = LDB:NewDataObject("SavedInstances", {
			label = "SavedInstances",
			type = "launcher",
			icon = "Interface\\Addons\\SavedInstances\\icon.tga",
			OnClick = function(clickedframe, button)
				self:ShowConfig()
			end,
			OnEnter = function(tt)
				SavedInstances.tooltip = QTip:Acquire("SavedInstancesTooltip", 1, "LEFT")
				SavedInstances.tooltip:SmartAnchorTo(tt)
				SavedInstances:ShowTooltip()
			end,
			OnLeave = SavedInstances.HideTooltip
		})
	end
	if LDBIcon then
		LDBIcon:Register("SavedInstances", self.launcher, self.db.MinimapIcon)
	end
	if LibStub:GetLibrary("LibFuBarPlugin-3.0", true) and not IsAddOnLoaded("FuBar2Broker") then
		local LFBP = LibStub:GetLibrary("LibFuBarPlugin-3.0")
		LibStub("AceAddon-3.0"):EmbedLibrary(self, "LibFuBarPlugin-3.0")
		self:SetFuBarOption("tooltipType", "Custom")
		self:SetFuBarOption("hasNoColor", true)
		self:SetFuBarOption("cannotDetachTooltip", true)
		self:SetFuBarOption("hideWithoutStandby", true)
		self:SetFuBarOption("iconPath", "Interface\\Addons\\SavedInstances\\icon.tga")
		self:SetFuBarOption("hasIcon", true)
		self:SetFuBarOption("defaultPosition", "RIGHT")
		self:SetFuBarOption("tooltipHiddenWhenEmpty", false)
		self:SetFuBarOption("configType", "None")
		LFBP:OnEmbedInitialize(self)
		self:GetFrame():SetScript("OnEnter", function()
			SavedInstances.tooltip = QTip:Acquire("SavedInstancesTooltip", 1, "LEFT")
			SavedInstances.tooltip:SmartAnchorTo(SavedInstances:GetFrame())
			SavedInstances:ShowTooltip()
		end)
		self:GetFrame():SetScript("OnLeave", SavedInstances.HideTooltip)
		function SavedInstances:OnFuBarClick(button)
				self:ShowConfig()
		end
		self:UpdateFuBarSettings()
		self.optionsFrames.FuBar = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SavedInstances", "FuBar Options", "SavedInstances", "FuBar")
	end
	self:RequestRefresh()
end

function SavedInstances:OnEnable()
	self:RegisterEvent("UPDATE_INSTANCE_INFO", "Refresh")
	self:RegisterEvent("RAID_INSTANCE_WELCOME", "RequestRefresh")
	self:RegisterEvent("CHAT_MSG_SYSTEM", "CheckSystemMessage")
end

function SavedInstances:OnDisable()
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")
end

function SavedInstances.HideTooltip()
	QTip:Release(SavedInstances.tooltip)
	SavedInstances.tooltip = nil
end

function SavedInstances:ShowTooltip()
	local tooltip = SavedInstances.tooltip
	local hFont = tooltip:GetHeaderFont()
	local hFontPath, hFontSize
	hFontPath, hFontSize, _ = hFont:GetFont()
	hFont:SetFont(hFontPath, hFontSize, "OUTLINE")
	tooltip:SetHeaderFont(hFont)
	self:CullStaleItems()
	local columns = { }
	local headLine, headCol = tooltip:AddHeader(NORMAL_FONT_COLOR_CODE.."SavedInstances"..FONT_COLOR_CODE_CLOSE)
	local i, instance
	for i, instance in ipairs(OrderedInstances()) do
		local currInstance = self.db.Instances[instance]
		if TableLen(currInstance) == 0 and self.db.Tooltip.AllRaids and currInstance.Show then
			tooltip:AddLine(GRAY_FONT_COLOR_CODE..currInstance..FONT_COLOR_CODE_CLOSE)
		else
			local currLine, _ = tooltip:AddLine(GRAY_FONT_COLOR_CODE..currInstance..FONT_COLOR_CODE_CLOSE)
			local toon, v
			for toon, v in pairs(currentInstance) do
			if not self.db.Toons[toon].Hide then
					tooltip:SetCell(currLine, 1, NORMAL_FONT_COLOR_CODE..instance..FONT_COLOR_CODE_CLOSE)
					if not columns[toon] then
						columns[toon] = tooltip:AddColumn("CENTER")
						tooltip:AddColumn("CENTER") 
					end	
					if instance[toon]["Normal"] then
						tooltip:SetCell(currLine, columns[toon], GREEN_FONT_COLOR_CODE.."N"..FONT_COLOR_CODE_CLOSE)
					end
					if instance[toon]["Heroic"] then
						tooltip:SetCell(currLine, columns[toon] + 1, RED_FONT_COLOR_CODE.."H"..FONT_COLOR_CODE_CLOSE)
					end
				end
			end
		end
	end
	if tooltip:GetLineCount() == 1 then
		tooltip:AddLine(GRAY_FONT_COLOR_CODE..NO_RAID_INSTANCES_SAVED..FONT_COLOR_CODE_CLOSE)
	end
	if self.db.Tooltip.FreshToons then
		local toon, class
		for toon, class in pairs(self.db.Class) do
			if not self.db.Toons[toon].Hide and not columns[toon] then
				columns[toon] = tooltip:AddColumn("CENTER")
				tooltip:AddColumn("CENTER")
			end
		end
	end
	local toon, col
	for toon, col in pairs(columns) do
		tooltip:SetCell(headLine, col, self:ClassColorise(self.db.Class[toon], strsplit(' ', toon)), tooltip:GetHeaderFont(), "CENTER", 2)
	end 
	local hintLine, hintCol = tooltip:AddLine()
	tooltip:SetCell(hintLine, hintCol, "|cffffff00".."Click|r to configure SavedInstances", "LEFT", tooltip:GetColumnCount()) 
	tooltip:Show()
end

function SavedInstances:CheckSystemMessage()
	if tostring(arg1) == INSTANCE_SAVED then
		self:RequestRefresh()
	end
end

function SavedInstances:RequestRefresh()
	RequestRaidInfo()
end

function SavedInstances:Refresh()
	local toon = UnitName("player").." - "..GetRealmName()
	self:CullStaleItems()
	self.db.NumSaved[toon] = GetNumSavedInstances()
	if self.db.NumSaved[toon] > 0 then
		local i, iname, iid, iexpires, idiff, temp
		for i = 1, self.db.NumSaved[toon] do
			iname, iid, iexpires, idiff = GetSavedInstanceInfo(i)
			if idiff > 1 then
				idiff = "Heroic"
			else
				idiff = "Normal"
			end
			self.db.Instances[iname] = self.db.Instances[iname] or { }
			self.db.Instances[iname].Order = self.db.Instances[iname].Order or LastInstance() + 1
			self.db.Instances[iname][toon] = self.db.Instances[iname][toon] or { }
			self.db.Instances[iname][toon][idiff] = {
				ID = iid,
				Expires = iexpires + time()
			}
		end
	end
	self:UpdateFeed()
end

function SavedInstances:UpdateFeed()
end

function SavedInstances:CullStaleItems()
	local instance, a, toon, b, difficulty, c
	for instance, a in pairs(SavedInstances.db.Instances) do
		for toon, b in pairs(a) do
			for difficulty, c in pairs (b) do
				if c.Expires < time() then
					c = nil
				end
			end
			if TableLen(b) == 0 then
				b = nil
			end
		end
	end
end

-- functions for maintaining the Instance order below

local function OrderedInstances()
	local list = { }
	local instance, a
	for instance, a in pairs(db.Instances) do
		list[a.Order] = instance
	end
	return list
end

local function LastInstance()
	local instance, a, last
	for instance, a in pairs(db.Instances) do
		last = last or a.Order
		if last < a.Order then
			last = a.Order
		end
	end
	last = last or 0
	return last
end

local function InstanceToFirst(tuple)
	if type(tuple) ~= "integer" then
		tuple = db.Instances[tuple].Order
	end
	local k, v
	for k, v in pairs(db.Instances) do
		if v.Order < tuple then
			v.Order = v.Order + 1
		end
	end
	db.Instances[tuple].Order = 1
end

local function InstanceToLast(tuple)
	local last = LastInstance()
	if type(tuple) ~= "integer" then
		tuple = db.Instances[tuple].Order
	end
	local k, v
	for k, v in pairs(db.Instances) do
		if v.Order > tuple then
			v.Order = v.Order - 1
		end
	end
	db.Instances[tuple].Order = last
end

-- options stuff below

function SavedInstances:UpdateFuBarSettings()
	if LibStub:GetLibrary("LibFuBarPlugin-3.0", true) then
		if SavedInstances.db.FuBar.HideMinimapButton then
			self:Hide()
		else
			self:Show()
			if self:IsFuBarMinimapAttached() ~= SavedInstances.db.FuBar.AttachMinimap then
				self:ToggleFuBarMinimapAttached()
			end
		end
	end
end

local function GetFuBarMinimapAttachedStatus(info)
	return SavedInstances:IsFuBarMinimapAttached() or SavedInstances.db.FuBar.HideMinimapButton
end

local options = {
	type = "group",
	name = "SavedInstances",
	handler = SavedInstances,
	args = {
		General = {
			order = 1,
			type = "group",
			name = L["General settings"],
			desc = L["General settings"],
			args = {
				intro = {
					order = 1,
					type = "description",
					name = L["Track the instance IDs saved against your characters"],
				},
				MinimapGroup = {
					order = 2,
					type = "group",
					guiInline = true,
					name = L["Minimap button"],
					desc = L["Minimap button"],
					hidden = function() return not LDBIcon end,
					args = {
						note = {
							order = 1,
							type = "description",
							name = L["Note: This is the recommended control for the minimap icon."],
						},
						MinimapIcon = {
							type = "toggle",
							name = L["Show minimap button"],
							desc = L["Show the SavedInstances minimap button"],
							order = 2,
							get = function(info) return not SavedInstances.db.MinimapIcon.hide end,
							set = function(info, value)
								SavedInstances.db.MinimapIcon.hide = not value
								if value then LDBIcon:Show("SavedInstances") else LDBIcon:Hide("SavedInstances") end
							end,
						},
					},
				},
				TooltipGroup = {
					order = 2,
					type = "group",
					guiInline = true,
					name = L["Tooltip contents"],
					desc = L["Tooltip contents"],
					args = {
						FreshToons = {
							type = "toggle",
							name = L["Show non-saved characters"],
							desc = L["Show characters that have not been saved to any instances"],
							order = 2,
							get = function(value)
									return SavedInstances.db.Tooltip.FreshToons
							end,
							set = function(info, value)
									SavedInstances.db.Tooltip.FreshToons = value
							end,
						},
						AllRaids = {
							type = "toggle",
							name = L["Show all current dungeons"],
							desc = L["Always show all dungeons from this expansion, even if none of your characters are saved to them"],
							order = 3,
							get = function(value)
									return SavedInstances.db.Tooltip.AllRaids
							end,
							set = function(info, value)
									SavedInstances.db.Tooltip.AllRaids = value
							end,
						},
					},
				},
			},
		},
		Ignore = {
			order = 4,
			type = "group",
			name = IGNORE_LIST,
			desc = L["Select which detected characters to never show in the tooltip"],
			args = {
				intro = {
					order = 1,
					type = "description",
					name = L["Select which detected characters to never show in the tooltip."],
				},
				List = {
					order = 2,
					type = "multiselect",
					name = L["Ignore these characters"],
--					desc = "Never show these characters in the tooltip",
					get = function(info, toon)
						return SavedInstances.db.Toons[toon].Hide
					end,
					set = function(info, toon, value)
						SavedInstances.db.Toons[toon].Hide = value
					end,
					values = function()
						local list = { }
						local k, v
						for k, v in pairs(SavedInstances.db.Toons) do
							list[k] = k
						end
						return list
					end,
				},
			},
		},
		FuBar = {
			order = -1,
			type = "group",
			name = "FuBar Options",
			desc = "FuBar Options",
			hidden = function() return SavedInstances.IsFuBarMinimapAttached == nil end,
			args = {
				intro = {
					order = 1,
					type = "description",
					name = "Configure FuBar and related options",
				},
				hideIcon = {
					type = "toggle",
					order = 2,
					name = "Hide minimap/FuBar icon",
					desc = "Hide minimap/FuBar icon",
					get = function(info) return SavedInstances.db.FuBar.HideMinimapButton end,
					set = function(info, v)
						SavedInstances.db.FuBar.HideMinimapButton = v
						SavedInstances:UpdateFuBarSettings()
					end,
				},
				showIcon = {
					type = "toggle",
					order = 3,
					name = "Show icon",
					desc = "Show icon",
					get = function(info) return SavedInstances:IsFuBarIconShown() end,
					set = function(info, v) SavedInstances:ToggleFuBarIconShown() end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				showText = {
					type = "toggle",
					order = 4,
					name = "Show text",
					desc = "Show text",
					get = function(info) return SavedInstances:IsFuBarTextShown() end,
					set = function(info, v) SavedInstances:ToggleFuBarTextShown() end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				position = {
					type = "select",
					order = 5,
					name = "Position",
					desc = "Position",
					values = {LEFT = "Left", CENTER = "Center", RIGHT = "Right"},
					get = function() return SavedInstances:GetPanel() and SavedInstances:GetPanel():GetPluginSide(SavedInstances) end,
					set = function(info, val)
						if SavedInstances:GetPanel() and SavedInstances:GetPanel().SetPluginSide then
							SavedInstances:GetPanel():SetPluginSide(SavedInstances, val)
						end
					end,
					disabled = GetFuBarMinimapAttachedStatus,
				},
				MinimapGroup = {
					order = 6,
					type = "group",
					guiInline = true,
					name = "Minimap button",
					desc = "Minimap button",
					args = {
						note = {
							order = 1,
							type = "description",
							name = "Note: It is recommended you leave the minimap control here disabled.",
						},
						attachMinimap = {
							type = "toggle",
							order = 2,
							name = "Attach to minimap",
							desc = "Attach to minimap",
							get = function(info) return SavedInstances:IsFuBarMinimapAttached() end,
							set = function(info, v)
								SavedInstances:ToggleFuBarMinimapAttached()
								SavedInstances.db.FuBar.AttachMinimap = SavedInstances:IsFuBarMinimapAttached()
							end,
							disabled = function() return SavedInstances.db.FuBar.HideMinimapButton end,
						},
					},
				},
			},
		},
	},
}
SavedInstances.Options = options

function SavedInstances:SetupOptions()
	self.optionsFrames = {}
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("SavedInstances", options)
	local ACD3 = LibStub("AceConfigDialog-3.0")
	self.optionsFrames.SavedInstances = ACD3:AddToBlizOptions("SavedInstances", nil, nil, "General")
	self.optionsFrames.Ignore = ACD3:AddToBlizOptions("SavedInstances", "Ignore Characters", "SavedInstances", "Ignore")
end

function SavedInstances:ShowConfig()
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.DateTime)
	InterfaceOptionsFrame_OpenToCategory(self.optionsFrames.SavedInstances)
end

-- general helper functions below

local function ClassColorise(class, targetstring)
	local class = strupper(gsub(class, " ", ""))
	local r = format("%x", math.floor(RAID_CLASS_COLORS[class].r * 255))
	if strlen(r) == 1 then
		r = "0"..r
	end
	local g = format("%x", math.floor(RAID_CLASS_COLORS[class].g * 255))
	if strlen(g) == 1 then
		g = "0"..g
	end
	local b = format("%x", math.floor(RAID_CLASS_COLORS[class].b * 255))
	if strlen(b) == 1 then
		b = "0"..b
	end
	return "|cff"..r..g..b..targetstring..FONT_COLOR_CODE_CLOSE
end

local function TableLen(table)
	local i = 0
	local k, v
	for k, v in pairs(table) do
		i = i + 1
	end
	return i
end
